import json
import random
from typing import List, TypedDict
from thefuzz import fuzz
from re import search


class Method(TypedDict):
    name: str
    arguments: List[str]


class Variable(TypedDict):
    name: str
    context_extractor: str
    method: Method
    fuzzy_extractor: List[str]


class Response(TypedDict):
    text: str
    variables: List[Variable]


class Trigger(Response):
    similarity: int


class Answer(TypedDict):
    name: str

    triggers: List[Trigger | str]
    responses: List[Response]


def addition(*args) -> str:
    return str(int(args[0]) + int(args[1]))


def subtraction(*args) -> str:
    return str(int(args[0]) - int(args[1]))


def multiplication(*args) -> str:
    return str(int(args[0]) * int(args[1]))


def division(*args) -> str:
    return str(int(args[0]) / int(args[1]))


def city_info(*args) -> str:
    city_def = {
        "Paris": "Paris, the 'City of Light,' is celebrated for its art, history, and gastronomy, featuring landmarks like the Eiffel Tower and Notre-Dame Cathedral. It's also a fashion and culinary capital, offering stylish boutiques and exquisite dining.",
        "Rome": "Rome, the 'Eternal City,' offers a journey through history with ancient ruins like the Colosseum and the Roman Forum, alongside Vatican City's marvels. Its vibrant streets are filled with delightful Italian cuisine.",
        "Prague": "Prague is renowned for its bridges, cathedrals, and gold-tipped towers reflecting in the Vltava River. The city's medieval center is a mix of cobbled lanes, walled courtyards, and church spires.",
        "Barcelona": "Barcelona is known for its art and architecture, highlighted by Gaudí's Sagrada Família and other modernist landmarks. The lively Las Ramblas street offers street performances and markets.",
        "London": "London is a blend of ancient history and modernity, home to iconic landmarks like the British Museum and Buckingham Palace. This bustling metropolis is also known for its diverse culinary scene and vibrant arts culture.",
        "Amsterdam": "Amsterdam is famous for its artistic heritage, elaborate canal system, and narrow houses with gabled facades. The city's Museum District houses works by Van Gogh and Rembrandt.",
        "Berlin": "Berlin, known for its modern architecture, turbulent history, and vibrant culture, is a city of contrast. Key sights include the Berlin Wall remnants and the Brandenburg Gate.",
        "Vienna": "Vienna is celebrated for its classical music heritage, imperial palaces, and rich artistic and intellectual legacy. The city's coffee-house culture and vibrant contemporary art scene are also notable",
        "Venice": "Venice is renowned for its beautiful waterways, Gothic and Renaissance architecture, and artworks. This unique city built on a lagoon is famous for its gondola rides and annual film festival.",
        "Dublin": "Dublin, known for its historical landmarks like Dublin Castle and St. Patrick's Cathedral, is rich in tradition and literary history. The city also boasts a lively pub scene and beautiful green spaces.",
        "Edinburgh": "Edinburgh is a city of striking contrasts, blending medieval relics in Old Town with the elegant Georgian New Town. The city is known for its historic castle and vibrant cultural festivals.",
        "Athens": "Athens is a historical powerhouse, home to ancient landmarks like the Acropolis and Parthenon temple. This city is a fascinating mix of classical history and modern life.",
        "Budapest": "Budapest is known for its dramatic cityscape marked by the Danube River, historic thermal baths, and architectural landmarks. The city's nightlife and cultural scene are vibrant and diverse.",
        "Lisbon": "Lisbon, perched on seven hills, is famous for its pastel-colored buildings and sunny climate. The city's rich history is evident in its imposing castle and picturesque old town.",
    }

    city = args[0]
    return city_def.get(city, "I don't know this city")


class Bot:
    answers: List[Answer]
    errors: List[str]

    def __init__(self, name: str):
        self.name = name

        self.ctx = {"name": self.name}
        self.methods = {
            "addition": addition,
            "subtraction": subtraction,
            "multiplication": multiplication,
            "division": division,
            "city_info": city_info,
        }

        self.load_answers("answers.json")

    def load_answers(self, filename):
        with open(filename, "r", encoding="utf-8") as file:
            data = file.read()
            json_data = json.loads(data)

            self.answers = json_data["answers"]
            self.errors = json_data["errors"]

    def find_answer(self, text: str) -> Answer | None:
        max_match = 50
        best_answer = None

        for answer in self.answers:
            for trigger in answer["triggers"]:
                if isinstance(trigger, str):
                    similarity = fuzz.token_set_ratio(text.lower(), trigger.lower())
                else:
                    trigger_text = self.process_variables(
                        text, trigger["text"], trigger.get("variables", [])
                    )

                    similarity = fuzz.token_sort_ratio(
                        text.lower(), trigger_text.lower(), full_process=True
                    )

                required_similarity = (
                    trigger.get("similarity", 50) if isinstance(trigger, dict) else 0
                )

                if similarity > max_match and similarity > required_similarity:
                    max_match = similarity
                    best_answer = answer

        return best_answer

    def process_answer(self, question: str, answer: Answer) -> str:
        response: Response = random.choice(answer["responses"])
        text = self.process_variables(
            question, response["text"], response.get("variables", [])
        )

        return text

    def process_variables(
        self, question: str, text: str, variables: List[Variable]
    ) -> str:
        for variable in variables:
            text = self.process_variable(question, text, variable)

        return text

    def process_variable(self, question: str, text: str, variable: Variable) -> str:
        context_extractor = variable.get("context_extractor", None)
        method = variable.get("method", None)
        fuzzy_extractor = variable.get("fuzzy_extractor", None)

        value = None

        if context_extractor:
            match = search(context_extractor, question)

            if match:
                value = match.group(1)

            self.ctx[variable["name"]] = f"{value if value else ''}"
        elif method:
            arguments = [self.ctx.get(arg, "") for arg in method["arguments"]]
            value = self.methods[method["name"]](*arguments)
            self.ctx[variable["name"]] = value
        elif fuzzy_extractor:
            [value, real_value] = self.fuzzy_extractor(question, fuzzy_extractor)
            self.ctx[variable["name"]] = value
            self.ctx[variable["name"] + "_real"] = real_value
        else:
            value = self.ctx.get(variable["name"], None)

        return text.replace(f"{{{variable['name']}}}", value if value else "")

    def fuzzy_extractor(self, text: str, true_values: List[str]) -> tuple[str, str]:
        words = text.split(" ")

        max_match = 50
        best_match = None
        real_value = None

        for word in words:
            for true_value in true_values:
                match = fuzz.token_sort_ratio(word.lower(), true_value.lower())

                if match > max_match:
                    max_match = match
                    best_match = true_value
                    real_value = word

        return (best_match if best_match else "", real_value if real_value else "")

    def chat(self):
        while True:
            try:
                question = input("You# ")
                answer = self.find_answer(question)

                if answer:
                    print(f"{self.name}#", self.process_answer(question, answer))

                    if answer["name"] == "exit":
                        break
                else:
                    raise Exception()
            except KeyboardInterrupt:
                print("\nGoodbye!")
                break
            except Exception as e:
                print(random.choice(self.errors))


def main():
    bot = Bot("Ravin")
    bot.chat()


if __name__ == "__main__":
    main()
